package flink.zanxiangnet.ad.monitoring;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import com.tencent.ads.model.DailyReportsGetListStruct;
import com.tencent.ads.model.HourlyReportsGetListStruct;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.maxcompute.sink.TunnelBatchSink;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfDayDTO;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfMinuteDTO;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdStatOfDayODSDTO;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.pojo.properties.KafkaProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.SerializableTimestampAssigner;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.functions.KeySelector;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.OffsetResetStrategy;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.config.SslConfigs;
import org.springframework.beans.BeanUtils;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

public class PlanStatJob {

    /**
     * 可能有数据的最早日期
     */
    private static final String OLDEST_DAY = "2019-01-01";

    public static void main(String[] args) throws Exception {
        System.setProperty("javax.net.ssl.trustStore", "D:\\Downloads\\kafka.client.truststore.jks");
        // System.setProperty("javax.net.ssl.trustStore", "/root/flink-1.13.2/kafka.client.truststore.jks");
        System.setProperty("javax.net.ssl.trustStorePassword", "KafkaOnsClient");
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(PlanStatJob.class.getResourceAsStream("/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        Properties adStreamOfMinuteProps = new Properties();
        adStreamOfMinuteProps.load(PlanStatJob.class.getResourceAsStream("/ad_stream_of_minute.properties"));
        KafkaSource<String> adStreamOfMinuteSource = buildKafkaSource(adStreamOfMinuteProps);

        DataStreamSource<String> adStreamOfMinuteIn = env.fromSource(adStreamOfMinuteSource, WatermarkStrategy.noWatermarks(), "adStreamOfMinuteSource_kafka");

        // 广告分钟数据（前 5分钟的广告消耗数据）
        final OutputTag<AdDataOfMinuteODS> adMinuteStreamTag = new OutputTag<AdDataOfMinuteODS>("adMinuteStream") {
        };
        // 广告小时数据（往前回滚 10天）
        final OutputTag<AdDataOfHourODS> adHourStreamTag = new OutputTag<AdDataOfHourODS>("adHourStream") {
        };

        // 对流进行映射，拆分（实时的分钟流和回滚的小时流）
        SingleOutputStreamOperator<AdDataOfMinuteODS> adODSStream = adStreamOfMinuteIn.filter(StringUtils::isNotBlank)
                .process(new ProcessFunction<String, AdDataOfMinuteODS>() {
                    @Override
                    public void processElement(String adDataStr, ProcessFunction<String, AdDataOfMinuteODS>.Context context, Collector<AdDataOfMinuteODS> collector) throws Exception {
                        AdDataOfMinuteDTO dto = JsonUtil.toObj(adDataStr, AdDataOfMinuteDTO.class);
                        // 指记录被创建的时间
                        long createTime = dto.getCreateTime();
                        // 指记录统计的时间（如果是实时统计的数据精确到分钟，回滚的历史数据精确到天）
                        long statTime = DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(dto.getDataTime()).withSecond(0).withNano(0));
                        HourlyReportsGetListStruct struct = dto.getHourlyReportsGetListStruct();

                        if (createTime - statTime > (60 * 60 * 1000L)) {
                            AdDataOfHourODS adODS = new AdDataOfHourODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(statTime).withMinute(0)));
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            context.output(adHourStreamTag, adODS);
                        } else {
                            AdDataOfMinuteODS adODS = new AdDataOfMinuteODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(statTime);
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            context.output(adMinuteStreamTag, adODS);
                        }
                    }
                });

        // 分钟流
        DataStream<AdDataOfMinuteODS> adMinuteODSStream = adODSStream.getSideOutput(adMinuteStreamTag);
        // 写入原始表
        adMinuteODSStream.addSink(new TunnelBatchSink<>(AdDataOfMinuteODS.class, 36000L, 64000L, 3));
        adMinuteODSStream
                // 打水印，允许数据延迟 6分钟，同时指定时间流
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfMinuteODS>forBoundedOutOfOrderness(Duration.ofMinutes(6L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfMinuteODS>) (adODS, l) -> adODS.getStatTime()))
                .keyBy(AdDataOfMinuteODS::getAdId)
                // 开一个 5分钟的滚动窗口
                .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                .process(new ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>() {
                    // 对象锁，防止MaxCompute的 Tunnel对象多次初始化
                    private final String DUMMY_LOCK = "DUMMY_LOCK";
                    private Odps odps;
                    private long lastClearTime = System.currentTimeMillis();
                    private final Map<Long, String> lastAdDayQueryDay = new ConcurrentHashMap<>();
                    // 历史的天数据
                    private final Map<Long, AdStatOfDayDWD> historyDayData = new ConcurrentHashMap<>();
                    // 之前聚合的昨天的数据
                    private final Map<Long, Map<String, AdStatOfMinuteDWD>> historyMinuteData = new ConcurrentHashMap<>();
                    // 前 5分钟聚合的数据
                    private final Map<Long, Map<String, AdStatOfMinuteDWD>> lastReduceData = new ConcurrentHashMap<>();

                    private void clearData() {
                        lastAdDayQueryDay.entrySet().removeIf(entry -> DateUtil.parseLocalDate(entry.getValue()).compareTo(LocalDate.now().minusDays(3L)) <= 0);
                        historyDayData.entrySet().removeIf(entry -> DateUtil.parseLocalDate(entry.getValue().getStatDay()).compareTo(LocalDate.now().minusDays(50L)) <= 0);
                        historyMinuteData.forEach((key, value) -> {
                            if (value != null) {
                                value.entrySet().removeIf(entry2 -> DateUtil.parseLocalDate(entry2.getKey()).compareTo(LocalDate.now().minusDays(3L)) <= 0);
                            }
                        });
                        historyMinuteData.entrySet().removeIf(entry -> entry.getValue() == null || entry.getValue().isEmpty());

                        lastReduceData.forEach((key, value) -> {
                            if (value != null) {
                                value.entrySet().removeIf(entry -> DateUtil.parseOfMinuteReduce(entry.getKey()).compareTo(LocalDateTime.now().minusHours(1)) <= 0);
                            }
                        });
                        lastReduceData.entrySet().removeIf(entry -> entry.getValue() == null || entry.getValue().isEmpty());
                    }

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
                    }

                    @Override
                    public void process(Long elementCount, ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context,
                                        Iterable<AdDataOfMinuteODS> iterable, Collector<AdStatOfMinuteDWD> collector) throws Exception {
                        LocalDateTime beginTime = DateUtil.milliToLocalDateTime(context.window().getStart());
                        String statDay = DateUtil.formatLocalDate(beginTime.toLocalDate());
                        String reduceTimeKey = DateUtil.formatOfMinuteReduce(beginTime);
                        int hour = beginTime.getHour();
                        long now = System.currentTimeMillis();
                        AdDataOfMinuteODS statODS = null;
                        List<AdDataOfMinuteODS> adDataOfMinuteODSList = new ArrayList<>(24);
                        for (AdDataOfMinuteODS adDataOfMinuteODS : iterable) {
                            adDataOfMinuteODSList.add(adDataOfMinuteODS);
                            if (adDataOfMinuteODS.getHour() != hour) {
                                continue;
                            }
                            statODS = adDataOfMinuteODS;
                            break;
                        }
                        if (statODS == null) {
                            return;
                        }

                        // 每 12小时清理一次数据
                        if (now - lastClearTime > 12 * 60 * 60 * 1000) {
                            synchronized (DUMMY_LOCK) {
                                if (now - lastClearTime > 12 * 60 * 60 * 1000) {
                                    clearData();
                                    lastClearTime = now;
                                }
                            }
                        }
                        Long adId = statODS.getAdId();

                        // 昨天聚合的数据
                        Map<String, AdStatOfMinuteDWD> historyMinuteMapping = historyMinuteData.computeIfAbsent(adId, key -> new HashMap<>());
                        AdStatOfMinuteDWD yesterdayMinuteDWD = historyMinuteMapping.get(DateUtil.formatLocalDate(beginTime.minusDays(1L).toLocalDate()));
                        // 之前的数据
                        String lastQueryDay = lastAdDayQueryDay.get(adId);
                        if (lastQueryDay == null || !lastQueryDay.equals(statDay)) {
                            // 往前找 30天
                            LocalDate endDay = beginTime.minusDays(2L).toLocalDate(), beginDay = endDay.minusDays(30L);
                            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + endDay + "\" AND ad_id = " + adId + ";");
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            List<AdStatOfDayDWD> temp = records.stream().map(AdStatOfDayDWD::byMaxCompute).sorted((val1, val2) -> {
                                if (val1.getStatDay().equals(val2.getStatDay())) {
                                    return new Long(val1.getCreateTime().getTime() - val2.getCreateTime().getTime()).intValue();
                                }
                                return DateUtil.parseLocalDate(val1.getStatDay()).compareTo(DateUtil.parseLocalDate(val2.getStatDay()));
                            }).collect(Collectors.toList());
                            historyDayData.put(adId, temp.isEmpty() ? null : temp.get(temp.size() - 1));
                            lastAdDayQueryDay.put(adId, statDay);
                        }
                        AdStatOfDayDWD beforeYesterdayDayDWD = historyDayData.get(adId);
                        Map<String, AdStatOfMinuteDWD> lastReduceMapping = lastReduceData.computeIfAbsent(adId, key -> new HashMap<>());
                        // 聚合当天的全部数据
                        AdStatOfMinuteDWD nowAdStat = AdStatOfMinuteDWD.reduce(beforeYesterdayDayDWD, yesterdayMinuteDWD, adDataOfMinuteODSList, statODS, lastReduceMapping.get(DateUtil.formatOfMinuteReduce(beginTime.minusMinutes(5L))), now);
                        collector.collect(nowAdStat);
                        lastReduceMapping.put(reduceTimeKey, nowAdStat);
                        historyMinuteMapping.put(statDay, nowAdStat);
                    }
                }).addSink(new TunnelBatchSink<>(AdStatOfMinuteDWD.class, 30000L, 365L, 6));
        ;

        // 小时流（直接写到小时报表的 ods）
        DataStream<AdDataOfHourODS> adHourODSStream = adODSStream.getSideOutput(adHourStreamTag);
        // 写入原始表
        adHourODSStream.addSink(new TunnelBatchSink<>(AdDataOfHourODS.class, 36000L, 64000L, 10));
        adHourODSStream.keyBy(AdDataOfHourODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    private final Map<Long, Long> lastAdQueryTime = new ConcurrentHashMap<>();
                    private final Map<Long, Map<String, Map<Integer, AdStatOfHourDWD>>> historyAdData = new ConcurrentHashMap<>();

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfHourODS> iterable, Collector<AdStatOfHourDWD> collector) throws Exception {
                        AdDataOfHourODS element = iterable.iterator().next();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        Integer hour = element.getHour();
                        long now = System.currentTimeMillis();
                        Long adId = element.getAdId();

                        Map<String, Map<Integer, AdStatOfHourDWD>> historyData = historyAdData.computeIfAbsent(adId, key -> new HashMap<>());
                        Long lastQueryTime = lastAdQueryTime.get(adId);
                        // 从 maxCompute拉取指定 广告的历史数据
                        if (lastQueryTime == null || (now - lastQueryTime) > 3 * 60 * 60 * 1000) {
                            LocalDate endTime = LocalDate.now(), beginTime = statDay.minusDays(40);
                            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_hour_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginTime) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endTime) + "\" AND ad_id = " + element.getAdId() + ";");
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            List<AdStatOfHourDWD> historyList = records.stream()
                                    .map(AdStatOfHourDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                                    .collect(Collectors.toList());
                            for (AdStatOfHourDWD adStatOfHourDWD : historyList) {
                                Map<Integer, AdStatOfHourDWD> hourMapping = historyData.computeIfAbsent(adStatOfHourDWD.getStatDay(), key -> new HashMap<>());
                                hourMapping.put(adStatOfHourDWD.getHour(), adStatOfHourDWD);
                            }
                            lastAdQueryTime.put(adId, now);
                        }
                        // 找到这个号当天内上个小时统计的数据
                        AdStatOfHourDWD lastHourData = null;
                        // 往前查 40天，找到广告上一次聚合的数据
                        for (int i = 0; i < 40; i++) {
                            Map<Integer, AdStatOfHourDWD> mapData = historyData.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
                            if (mapData == null || mapData.isEmpty()) {
                                continue;
                            }
                            for (Map.Entry<Integer, AdStatOfHourDWD> temp : mapData.entrySet()) {
                                if (temp.getKey() >= hour) {
                                    continue;
                                }
                                if (lastHourData != null && lastHourData.getHour() >= temp.getKey()) {
                                    continue;
                                }
                                lastHourData = temp.getValue();
                            }
                            if (lastHourData != null) {
                                break;
                            }
                        }
                        AdStatOfHourDWD newStatData = AdStatOfHourDWD.reduce(lastHourData, element, now);
                        Map<Integer, AdStatOfHourDWD> hourDataMapping = historyData.computeIfAbsent(newStatData.getStatDay(), key -> new HashMap<>());
                        hourDataMapping.put(newStatData.getHour(), newStatData);
                        collector.collect(newStatData);
                    }
                }).addSink(new TunnelBatchSink<>(AdStatOfHourDWD.class, 30000L, 365L, 6));

        // 获取指定广告的历史统计信息（用于统计总的消耗信息等）
        /*SingleOutputStreamOperator<AdStatOfHourDWD> adHourDWDStream = adHourODSStream.map(AdStatOfHourDWD::byODS)
                .keyBy((KeySelector<AdStatOfHourDWD, String>) adHourDWD -> adHourDWD.getStatDay() + "_" + adHourDWD.getAdId())
                // 开一个按天滚动的窗口
                .window(TumblingEventTimeWindows.of(Time.days(1L)))
                // 数据聚合
                .reduce(new RichReduceFunction<AdStatOfHourDWD>() {
                    // 初始化 MaxCompute连接实例
                    private Odps odps;

                    @Override
                    public void open(Configuration parameters) throws Exception {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        Odps odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
                    }

                    @Override
                    public AdStatOfHourDWD reduce(AdStatOfHourDWD adStatOfHourDWD, AdStatOfHourDWD t1) {
                        Long accountId = adStatOfHourDWD.getAccountId();
                        Long campaignId = adStatOfHourDWD.getCampaignId();
                        Long adId = adStatOfHourDWD.getAdId();
                        String statDay = adStatOfHourDWD.getStatDay();
                        // 1、从 MaxCompute取出小时数据往前的所有无重复按天统计的数据，并聚合用来算总的
                        // 2、当前流的所有数据聚合在一起作为天数据
                        // 3、写出到 MaxCompute
                        return null;
                    }
                });*/


        // ------------------------------------------------------- 处理广告的天数据 -----------------------------------------
        Properties adStreamOfDayProps = new Properties();
        adStreamOfDayProps.load(PlanStatJob.class.getResourceAsStream("/ad_stream_of_day.properties"));
        KafkaSource<String> adStreamOfDaySource = buildKafkaSource(adStreamOfDayProps);

        DataStreamSource<String> adStreamOfDayIn = env.fromSource(adStreamOfDaySource, WatermarkStrategy.noWatermarks(), "adStreamOfMinuteSource_kafka");

        // 广告日数据。往前回滚 10天
        final OutputTag<AdDataOfDayODS> adDayStreamRollDayTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollDayTag") {
        };
        // 广告日数据。往前回滚 1年
        final OutputTag<AdDataOfDayODS> adDayStreamRollYearTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollYearTag") {
        };
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStream = adStreamOfDayIn.filter(StringUtils::isNotBlank)
                .map(str -> {
                    AdDataOfDayDTO dto = JsonUtil.toObj(str, AdDataOfDayDTO.class);
                    Date createTime = new Date(dto.getCreateTime());
                    DailyReportsGetListStruct struct = dto.getDailyReportsGetListStruct();
                    AdDataOfDayODS adODS = new AdDataOfDayODS();
                    BeanUtils.copyProperties(struct, adODS);
                    adODS.setStatDay(struct.getDate());
                    adODS.setAccountId(dto.getAccountId());
                    adODS.setCampaignId(struct.getCampaignId());
                    adODS.setAgencyAccountId(struct.getAccountId());
                    adODS.setWechatAccountId(struct.getWechatAccountId());
                    adODS.setAdgroupId(struct.getAdgroupId());
                    adODS.setAdId(struct.getAdId());
                    adODS.setCreateTime(createTime);
                    return AdStatOfDayODSDTO.builder()
                            .startDate(dto.getStartDate())
                            .endDate(dto.getEndDate())
                            .adDataOfDayODS(adODS)
                            .build();
                });
        adDayODSStream.map(AdStatOfDayODSDTO::getAdDataOfDayODS).addSink(new TunnelBatchSink<>(AdDataOfDayODS.class, 36000L, 6000L, 10));

        // 拆分流
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStreamSplit = adDayODSStream.process(new ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>() {
            @Override
            public void processElement(AdStatOfDayODSDTO adDataDTO, ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>.Context context, Collector<AdStatOfDayODSDTO> collector) throws Exception {
                LocalDate startDate = adDataDTO.getStartDate();
                LocalDate endDate = adDataDTO.getEndDate();
                AdDataOfDayODS adODS = adDataDTO.getAdDataOfDayODS();
                if (DateUtil.intervalOfDays(startDate, endDate) > 31L) {
                    // 拉取时间间隔超过 1个月，账号回滚 365天的数据
                    context.output(adDayStreamRollYearTag, adODS);
                } else {
                    // 每日往前回滚 10天的数据
                    context.output(adDayStreamRollDayTag, adODS);
                }
            }
        });

        adDayODSStreamSplit.getSideOutput(adDayStreamRollDayTag)
                // 打水印，允许延迟 10天（应为允许回滚 10天），同时指定事件时间
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfDayODS>forBoundedOutOfOrderness(Duration.ofDays(10L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfDayODS>) (adDay, l) -> DateUtil.localDateToMilli(DateUtil.parseLocalDate(adDay.getStatDay())))
                ).keyBy(AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    private final Map<Long, Long> lastAdQueryTime = new ConcurrentHashMap<>();
                    private final Map<Long, Map<String, AdStatOfDayDWD>> historyAdData = new ConcurrentHashMap<>();

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfDayODS> iterable, Collector<AdStatOfDayDWD> collector) throws Exception {
                        AdDataOfDayODS element = iterable.iterator().next();
                        Long adId = element.getAdId();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        long now = System.currentTimeMillis();

                        Map<String, AdStatOfDayDWD> historyData = historyAdData.computeIfAbsent(adId, key -> new HashMap<>());
                        Long lastQueryTime = lastAdQueryTime.get(adId);
                        // 从 maxCompute查找广告的历史数据
                        if (lastQueryTime == null || (now - lastQueryTime) > 60 * 60 * 1000) {
                            LocalDate endTime = LocalDate.now(), beginTime = statDay.minusDays(60);
                            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginTime) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endTime) + "\" AND ad_id = " + element.getAdId() + ";");
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            List<AdStatOfDayDWD> list = records.stream()
                                    .map(AdStatOfDayDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue()).collect(Collectors.toList());
                            for (AdStatOfDayDWD adStatOfDayDWD : list) {
                                historyData.put(adStatOfDayDWD.getStatDay(), adStatOfDayDWD);
                            }
                            lastAdQueryTime.put(adId, now);
                        }
                        AdStatOfDayDWD newStatData = null;
                        for (int i = 1; i <= 60; i++) {
                            AdStatOfDayDWD oldStatData = historyData.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
                            if (oldStatData == null) {
                                continue;
                            }
                            newStatData = AdStatOfDayDWD.reduce(oldStatData, element, now);
                            break;
                        }
                        if (newStatData == null) {
                            newStatData = AdStatOfDayDWD.reduce(null, element, now);
                        }
                        historyData.put(newStatData.getStatDay(), newStatData);
                        collector.collect(newStatData);
                    }
                }).addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));

        adDayODSStreamSplit.getSideOutput(adDayStreamRollYearTag)
                .keyBy((KeySelector<AdDataOfDayODS, Long>) AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    // 上次聚合的结果
                    private AdStatOfDayDWD lastReduce;

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfDayODS> elements, Collector<AdStatOfDayDWD> out) {
                        AdDataOfDayODS element = elements.iterator().next();
                        lastReduce = AdStatOfDayDWD.reduce(lastReduce, element, System.currentTimeMillis());
                        out.collect(lastReduce);
                    }
                }).addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));

        env.execute();
    }

    private static KafkaSource<String> buildKafkaSource(Properties props) {
        return buildKafkaSource(props.getProperty(KafkaProperties.KAFKA_SERVERS),
                props.getProperty(KafkaProperties.KAFKA_USERNAME),
                props.getProperty(KafkaProperties.KAFKA_PASSWORD),
                props.getProperty(KafkaProperties.KAFKA_SSL_PATH),
                props.getProperty(KafkaProperties.KAFKA_TOPIC),
                props.getProperty(KafkaProperties.KAFKA_GROUP_ID)
        );
    }

    /**
     * 内网
     *
     * @param servers
     * @param username
     * @param password
     * @param sslPath
     * @param topic
     * @param groupId
     * @return
     */
    private static KafkaSource<String> buildKafkaSourceVPC(String servers, String username, String password, String sslPath, String topic, String groupId) {
        Properties props = new Properties();
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, servers);
        //两次Poll之间的最大允许间隔。
        //消费者超过该值没有返回心跳，服务端判断消费者处于非存活状态，服务端将消费者从Group移除并触发Rebalance，默认30s。
        props.put(ConsumerConfig.SESSION_TIMEOUT_MS_CONFIG, 30000);
        //每次Poll的最大数量。
        //注意该值不要改得太大，如果Poll太多数据，而不能在下次Poll之前消费完，则会触发一次负载均衡，产生卡顿。
        props.put(ConsumerConfig.MAX_POLL_RECORDS_CONFIG, 30);
        //消息的反序列化方式。
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        //当前消费实例所属的消费组，请在控制台申请之后填写。
        //属于同一个组的消费实例，会负载消费消息。
        props.put(ConsumerConfig.GROUP_ID_CONFIG, groupId);
        return KafkaSource.<String>builder()
                .setProperties(props)
                .setBootstrapServers(servers)
                .setTopics(topic)
                .setGroupId(groupId)
                .setStartingOffsets(OffsetsInitializer.committedOffsets(OffsetResetStrategy.EARLIEST))
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }

    /**
     * 公网
     *
     * @param servers
     * @param username
     * @param password
     * @param sslPath
     * @param topic
     * @param groupId
     * @return
     */
    private static KafkaSource<String> buildKafkaSource(String servers, String username, String password, String sslPath, String topic, String groupId) {
        Properties props = new Properties();
        props.put(SaslConfigs.SASL_JAAS_CONFIG, "org.apache.kafka.common.security.plain.PlainLoginModule required username=\"" + username + "\" password=\"" + password + "\";");
        props.put(SslConfigs.SSL_TRUSTSTORE_LOCATION_CONFIG, sslPath);
        props.put(SslConfigs.SSL_TRUSTSTORE_PASSWORD_CONFIG, "KafkaOnsClient");
        props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_SSL");
        props.put(SaslConfigs.SASL_MECHANISM, "PLAIN");
        props.put(SslConfigs.SSL_ENDPOINT_IDENTIFICATION_ALGORITHM_CONFIG, "");
        props.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringSerializer");
        props.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringSerializer");
        return KafkaSource.<String>builder()
                .setBootstrapServers(servers)
                .setTopics(topic)
                .setGroupId(groupId)
                .setProperties(props)
                .setStartingOffsets(OffsetsInitializer.committedOffsets(OffsetResetStrategy.EARLIEST))
                .setValueOnlyDeserializer(new SimpleStringSchema())
                .build();
    }
}
